/*
 * HASH.C - Generic hash handler module.  Include HASH.H
 *
 * $Header:   C:/fe/shared/vcs/hash.c_v   1.4   04 Jun 1990 14:59:58   rji  $
 *
 * $Log:   C:/fe/shared/vcs/hash.c_v  $
 * 
 *    Rev 1.4   04 Jun 1990 14:59:58   rji
 * Added hdelete() function to remove a single entry from a hash table.
 * Function hcreate() now requires the specification of three new function
 * pointer parameters -- these permit the client to specify his or her own
 * randomizing ("hash") function, comparison function (for table matching)
 * and copy function (for controlling how a hash name value is saved in a
 * table entry).  Generalized the formal parameter types of some existing
 * functions.  Amplified comments.  Rearranged functions into alphabetical
 * order and other cosmetic improvements.
 * 
 *    Rev 1.3   03 Sep 1989 23:11:08   rji
 * Changed #include file specifiers from <...> to "..." for non-standard
 * headers.
 * 
 *    Rev 1.2   09 Aug 1989 17:09:26   rji
 * Function headers changed to new ANSI format.
 * 
 *    Rev 1.1   14 Feb 1989 12:33:42   rji
 * Changed register to REGi, as appropriate.
 * 
 *    Rev 1.0   07 Dec 1988 23:10:24   rji
 * Initial revision.
 * 
*/

/*=========================================================================*/
/*=                                                                       =*/
/*=      I N C L U D E S,  D E F I N E S,  D E C L A R A T I O N S        =*/
/*=                                                                       =*/
/*=========================================================================*/

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         HASH.C Includes

  $PATHS$
       Functions\C
       Shared\Includes
       Shared\Utils
       Modules\$./FL$

   $3$

   See Also:
        $/SEE(COMMON.H Macros)$
        $/SEE(DEBUG.H Macros)$
        $/SEE(DALLOC.H Macros)$
        $/SEE(QUEUE.H Macros)$
        $/SEE(HASH.H Macros)$

   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* INCLUDES                                                                */
/*-------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "common.h"
#include "debug.h"
#include "dalloc.h"
#include "queue.h"

#include "hash.h"
/*                                                                   $end$ */

#define  EOS         '\0'                       /* End-Of-String delimiter */

/*-------------------------------------------------------------------------*/
/* PRIVATE FUNCTIONS                                                       */
/*-------------------------------------------------------------------------*/

static size_t  default_rand( HASH, void *);
static int     default_comp( void *, void * );
static void    default_copy( void *, void *, size_t );
#if !defined(NDEBUG)
#define  hash_check(h,s)        hcheck((h),(s))
static void    hcheck( HASH, char * );
#else
#define  hash_check(h,s)
#endif

/*-------------------------------------------------------------------------*/
/* PRIVATE DATA                                                            */
/*-------------------------------------------------------------------------*/

/* The following static variable, curr_tbl, is set by functions that
** search hash tables, like hadd() and hfind(), to establish the currently
** active hash table.  default_comp() uses curr_tbl to retrieve the name size
** for the active table, so that comparisons can be limited to the maximum
** number of significant characters in a name for the active table.
** Since we chose to use the QUEUE functions as a basis for the HASH
** functions, the comparison function default_comp() used as an argument to
** queue_find() in hadd() and hfind() must take only the two objects to be
** compared as arguments.  Originally, default_comp() took the hash table as
** a third argument, allowing it access to the active name size in a
** reentrant manner.
*/

static HASH       curr_tbl;

/*=========================================================================*/
/*=                                                                       =*/
/*=                F U N C T I O N   D E F I N I T I O N S                =*/
/*=                                                                       =*/
/*=========================================================================*/

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         default_comp()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* DEFAULT_COMP                                                            */
/*                                                                         */
/* Default name matching function.  Called by hadd() and hfind(), if       */
/* client does not supply his own via hcreate(), to locate the table entry */
/* whose name matches a given name.  Assumes both search key and names in  */
/* the hash table are null-terminated character strings.  Only the first   */
/* name_size - 1 characters are compared, since name_size for strings must */
/* include space for the terminal null byte.                               */
/*                                                                         */
/* Returns 0 if and only if match succeeds.                                */
/*-------------------------------------------------------------------------*/

static int  default_comp(
   void  *key,                   /* in: name to compare with hash element  */
   void  *ele                    /* in: hash element to compare with name  */
)
/*                                                                   $end$ */
{
   return ( strnicmp( (char *)key, (char *)ele, curr_tbl->name_size-1 ) );

} /* default_comp() */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         default_copy()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* DEFAULT_COPY                                                            */
/*                                                                         */
/* Default name copying function.  Called by hadd(), if client does not    */
/* supply his own via hcreate(),  to copy name to a newly allocated hash   */
/* table entry.  Assumes name is a null-terminated character string.  Only */
/* the first name_size - 1 characters are copied into a table entry, since */
/* name_size for strings must include space for the terminal null byte.    */
/* A null byte is guaranteed to terminate the name as stored.              */
/*-------------------------------------------------------------------------*/

static void default_copy(
   void    *tbl_name,               /* out: where to store name in table   */
   void    *name,                   /* in:  name to be stored in table     */
   size_t   tbl_name_size           /* in:  name_size of table             */
)
/*                                                                   $end$ */
{
   strscpy( (char *)tbl_name, (char *)name, tbl_name_size );

} /* default_copy() */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         default_rand()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* DEFAULT_RAND                                                            */
/*                                                                         */
/* Default randomizing function.  Called by hadd() and hfind(), if client  */
/* does not supply his own via hcreate(), to derive the hash table index   */
/* for a name.  Assumes name is a null-terminated character string.  Only  */
/* the first name_size - 1 characters used to derive the index, since      */
/* name_size for strings must include space for the terminal null byte.    */
/*                                                                         */
/* Returns hash table index derived from the given name.                   */
/*-------------------------------------------------------------------------*/

static size_t  default_rand(
   HASH  h,                         /* in: hash table to hash name into    */
   void *name                       /* in: name to be hashed into table    */
)
/*                                                                   $end$ */
{
   REG1 size_t i;
   REG2 size_t sum = 0;

   for ( i = 0; *((char *)name+i) != EOS  &&  i < h->name_size-1; i++ )
      sum += toupper( *((char *)name+i) );

#if defined(DRIVER)
   printf( "default_rand( %s ) = %u\n", name, sum % h->table_size );
#endif

   return( sum % h->table_size );

} /* default_rand() */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         hadd()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* HADD                                                                    */
/*                                                                         */
/* Adds element to hash. Caller supplies size of storage required to hold  */
/* hash element information, and HADD allocates that amount plus whatever  */
/* storage is needed to maintain hash structure.  HADD returns a pointer   */
/* to an area the caller can then fill in for the particular application   */
/* purpose at hand.  If a hash element already exists for the given name,  */
/* no storage is allocated; rather, a pointer to the existing element is   */
/* returned.                                                               */
/*-------------------------------------------------------------------------*/

void  *hadd(
   HASH     h,          /* in/out: hash to which element is to be added    */
   void    *nm,         /* in: name under which to store element data      */
   size_t   es          /* in: size of element to be added to hash         */
)
/*                                                                   $end$ */
{
   size_t   index;
   char    *p;

   hash_check( h, "hadd" );               /* valid hash? */

   curr_tbl = h;                          /* establish active hash table   */

   index = (*h->rand)( h, nm );           /* hash name to bucket index     */

   if ( (h->table)[index] == (QUEUE) NULL )
      /* if necessary, set up bucket */
      (h->table)[index] = queue_create();
   else
      /* bucket set up: if not empty & name found, return ptr to user area */
      if (  !queue_isempty( (h->table)[index] )
         && (p = queue_find( (h->table)[index], nm, (*h->comp) )) != NULL )
         return ( (void *) (p + h->name_size) );

   /* at this point bucket is set up and name was not found */

   p = queue_add( (h->table)[index], h->name_size + es );
   (*h->copy)( p, nm, h->name_size );
   ++h->tally;                            /* increment hash count */

   /* return pointer to user area */

   return ( (void *) (p + h->name_size) );

} /* hadd() */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         hcheck()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
#if !defined(NDEBUG)
/*-------------------------------------------------------------------------*/
/* HCHECK                                                                  */
/*                                                                         */
/* Checks integrity of a hash; can be turned off by re-compiling with      */
/* NDEBUG defined on compiler command line.                                */
/*-------------------------------------------------------------------------*/

static void hcheck(
   REG1 HASH   h,       /* in: hash whose integrity is to be checked       */
   char       *s        /* in: string to display if integrity check fails  */
)
/*                                                                   $end$ */
{
   REG2 int i;
   REG3 int hsum = 0;
   int      bad_hash = 0;

   if ( h == NULL )
      ++bad_hash;
   for ( i = 0; i < h->table_size; ++i )
      if ( h->table[i] != NULL )
         hsum += queue_length( h->table[i] );
   if ( hsum != h->tally )
      ++bad_hash;

   if ( bad_hash != 0 )
   {
      fprintf( stderr, "hcheck: messed up hash (%s)\n", s );
      exit( 1 );
   }

   return;

} /* hcheck() */

#endif

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         hclear()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* HCLEAR                                                                  */
/*                                                                         */
/* Removes all elements from a hash.                                       */
/*-------------------------------------------------------------------------*/

void   hclear(
   HASH  h          /* in: hash table to be cleared */
)
/*                                                                   $end$ */
{
   REG1 int i;

   hash_check( h, "hclear" );    /* valid hash? */

   /* free all hash elements */

   for ( i = 0; i < h->table_size; ++i )
      if ( ((h)->table)[i] != (QUEUE) NULL )
         queue_destroy(  (h->table)[i] );

   /* reset hash control variables */

   h->tally = 0;

   return;

} /* hclear() */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         hcreate()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* HCREATE                                                                 */
/*                                                                         */
/* Creates a HASH (allocates and initializes a hash header structure and   */
/* associated table) of specified table size and name size, and with       */
/* specified functions for randomizing names into table indices, for name  */
/* comparison and for copying names into table entries.  If any of the     */
/* function parameter values is NULL, the corresponding default function   */
/* will be used.                                                           */
/*                                                                         */
/* Table sizes should be prime numbers -- if names are character strings,  */
/* avoid primes of form R**k +/- a, where R is the radix of the character  */
/* code (e.g., 128, for ASCII) and k and a are small integers.             */
/* For further details on hash tables and functions, see Vol. 3 of Knuth.  */
/*                                                                         */
/* Returns the newly created HASH (i.e., a pointer to the hash header      */
/* structure it allocates.)                                                */
/*-------------------------------------------------------------------------*/

HASH  hcreate(
   size_t   tbl_sze,                            /* in: hash table size     */
   size_t   nme_sze,                            /* in: maximum name size   */
   size_t   (*rand)( HASH, void * ),            /* in: randomizing func    */
   int      (*comp)( void *, void * ),          /* in: name comparison func*/
   void     (*copy)( void *, void *, size_t )   /* in: name copying func   */
)
/*                                                                   $end$ */
{
   HASH     h;
   REG1 int i;

   if ( (h = (HASH) d_malloc( sizeof(HASH_HDR) + sizeof(QUEUE)*tbl_sze ))
         == NULL )
   {
      D( fprintf( stderr, "hcreate: no space for hash\n" ); )
      return ( NULL );
   }

   h->table       = (QUEUE *) ((char *) h + sizeof(HASH_HDR));
   h->table_size  = tbl_sze;
   h->tally       = 0;
   h->name_size   = nme_sze;
   h->rand        = (rand == (size_t (*)( HASH, void * )) NULL)
                    ? default_rand
                    : rand;
   h->comp        = (comp == (int (*)( void *, void * )) NULL)
                    ? default_comp
                    : comp;
   h->copy        = (copy == (void (*)( void *, void *, size_t )) NULL)
                    ? default_copy
                    : copy;

   for ( i = 0; i < tbl_sze; ++i )
      (h->table)[i] = NULL;

   return ( h );

} /* hcreate() */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         hdelete()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* HDELETE                                                                 */
/*                                                                         */
/* Deletes a hash element (frees element storage, adjusting hash structure */
/* as necessary).  Note that dynamically allocated storage attached to a   */
/* hash table element by the client program cannot be freed here.  The     */
/* client program must free such storage before deleting the table entry.  */
/*                                                                         */
/* Returns:                                                                */
/*             0 = element could not be deleted                            */
/*             1 = element successfully deleted                            */
/*-------------------------------------------------------------------------*/

int   hdelete(
   HASH     h,    /* in/out: hash from which element is to be deleted      */
   void   **epp   /* in/out: pointer to pointer to hash element to delete  */
)
/*                                                                   $end$ */
{
   QUEUE    q;
   size_t   index;
   char    *p;

   hash_check( h, "hdelete" );               /* valid hash? */

   /* if pointer to element pointer is invalid, return immediately */
   if ( epp == NULL  ||  *epp == NULL )
      return ( 0 );

   /* find bucket (queue) with element to be deleted */
   index = (*h->rand)( h, hash_name( h, *epp ) );
   q = (h->table)[index];

   /* quit if no bucket or bucket is empty */
   if ( q == NULL  ||  queue_isempty( q ) )
      return ( 0 );

   p = ((char *)(*epp) - h->name_size);      /* ptr to queue element */

   /* if element found, free element's storage and adjust hash */
   if ( queue_delete( q, p ) )
   {
      *epp = NULL;                     /* prevent access to freed storage  */
      --(h->tally);                                /* decrement hash count */
      return ( 1 );
   }
   else
      return ( 0 );

} /* hdelete() */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         hdestroy()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* HDESTROY                                                                */
/*                                                                         */
/* Destroys a hash by freeing all storage allocated to it -- by this       */
/* module!  Note that dynamically allocated storage attached to hash table */
/* elements by the client program cannot be freed here.  The client        */
/* program must sweep through the table, freeing all such storage, before  */
/* destroying the hash table.  The paired macros FOR_EACH_HASH_ENTRY and   */
/* END_FOR_HASH may be used to sweep through each table entry.                                                                        */
/*-------------------------------------------------------------------------*/

void  hdestroy(
   HASH *hp          /* in/out: pointer to hash (i.e., to hash header ptr) */
)
/*                                                                   $end$ */
{
   REG1 size_t i;

   hash_check( *hp, "hdestroy" );      /* valid hash? */

   /* free all hash elements */
   for ( i = 0; i < (*hp)->table_size; ++i )
      if ( ((*hp)->table)[i] != NULL )
         queue_destroy( ((*hp)->table)[i] );

   d_free( *hp );                      /* free hash header & table         */
   *hp = NULL;                         /* prevent access to freed storage  */

   return;

} /* hdestroy() */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         hdump()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* HDUMP                                                                   */
/*                                                                         */
/* Print hash table statistics.                                            */
/*-------------------------------------------------------------------------*/

void hdump(
   HASH  h           /* hash table to extract statistics from  */
)
/*                                                                   $end$ */
{
   REG1 size_t    i;
   REG2 size_t    max_bkt;
   REG3 size_t    min_bkt;
   REG4 size_t    buckets;

   buckets = max_bkt = 0;
   min_bkt = (size_t)(long) -1;        /* max 2's complement size_t value  */

   fprintf( stderr, "Hash Table Size: %u,  Maximum Name Size: %u\n",
                                      h->table_size,          h->name_size );

   for ( i = 0; i < h->table_size; ++i )
   {
      if ( h->table[i] != NULL )
      {
         ++buckets;
         max_bkt = max( max_bkt, queue_length( (h->table)[i]) );
         min_bkt = min( min_bkt, queue_length( (h->table)[i]) );
      }
   }

   fprintf( stderr, "Length of longest bucket: %u\n", max_bkt );
   fprintf( stderr, "Length of shortest bucket: %u\n", buckets ? min_bkt : 0 );
   fprintf( stderr, "Number of used buckets: %u\n"
                    "Total number of elements: %u\n",
                     buckets, h->tally );
   fprintf( stderr, "Average # elements/bucket: %.2f\n",
            buckets ? (double) h->tally / buckets : 0.0 );

} /* hdump() */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         hfind()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* HFIND                                                                   */
/*                                                                         */
/* Search for hash element and return a pointer to it, if found.  If it is */
/* not found, a NULL pointer is returned.                                  */
/*-------------------------------------------------------------------------*/

void *hfind(
   HASH  h,          /* in: hash table in which element is to be found     */
   void *nm          /* in: name to be searched for in hash table          */
)
/*                                                                   $end$ */
{
   size_t   index;
   char    *p;

   hash_check( h, "hfind" );              /* valid hash?                   */

   curr_tbl = h;                          /* establish active hash table   */

   index = (*h->rand)( h, nm );           /* hash name to bucket index     */

   if ( (h->table)[index] == NULL  )      /* no bucket ==> no match        */
      return ( NULL );

   /* search bucket for name match */

   if ( !queue_isempty( (h->table)[index] )
        && (p = queue_find( (h->table)[index], nm, (*h->comp) )) != NULL )
      return( (void *) (p + h->name_size) );

   /* if this point reached, no match found */

   return ( NULL );

} /* hfind() */



#if defined(DRIVER)
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         HASH.C main() Test Driver

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*=========================================================================*/
/*=                                                                       =*/
/*=                       T E S T   D R I V E R                           =*/
/*=                                                                       =*/
/*=========================================================================*/

typedef struct h_ele
{
   int name_type;
} h_ele;

void  main( int, char ** );

void main( int argc, char **argv )
/*                                                                   $end$ */
{
   char     s[30], *prog;
   HASH     h;
   int      i = 1, symbols = 0, misses = 0,
            add_prompt = 1, find_prompt = 1;
   size_t   tsize = 127, nsize = 12;
   h_ele    *hl;

   prog = argv[0];

   while ( --argc > 0  &&  (*++argv)[0] == '-' )
      switch ( (*argv)[1] )
      {
         case 'n':
            nsize = atoi( argv[0]+2 );
            break;
         case 't':
            tsize = atoi( argv[0]+2 );
            break;
         case '?':
         default:
            fprintf( stderr,
                     "usage: %s [-t<table_size>] [-n<name_size>] "
                     "[<add_file>] [<find_file>]\n", prog );
            exit( 1 );
      } /* switch */

   if ( argc-- > 0 )    /* open file of names to add, if specified   */
   {
      if ( ( freopen( *argv++, "r", stdin ) ) == NULL )
      {
         fprintf( stderr, "%s: cannot open file '%s'\n", prog, *--argv );
         exit( 1 );
      }
      else
         add_prompt = 0;
   }

   h = hash_create( tsize, nsize,
                    (size_t (*)( HASH, void * )) NULL,
                    (int (*)( void *, void * )) NULL,
                    (void (*)( void *, void *, size_t )) NULL );

   if ( add_prompt )
      printf( "Enter name, Ctrl-Z to end: " );

   while ( gets( s ) )
   {
      ++symbols;
      hl = hash_add( h, s, sizeof( h_ele ) );
      hl->name_type = i++;
      if ( add_prompt )
         printf( "Enter name, Ctrl-Z to end: " );
   }

   printf( "\nNumber of symbols read: %d\n\n", symbols );

   printf( "FOR_EACH_HASH_ENTRY element dump:\n\n" );
   FOR_EACH_HASH_ENTRY( h, hl, h_ele * )
      printf( "hl->name_type = %d\n", hl->name_type );
   END_FOR_HASH

   printf( "\n" );
   hash_dump( h );
   printf( "\n" );

   if ( argc-- > 0 )    /* open file of names to find, if specified   */
   {
      if ( ( freopen( *argv++, "r", stdin ) ) == NULL )
      {
         fprintf( stderr, "%s: cannot open file '%s'\n", prog, *--argv );
         exit( 1 );
      }
      else
         find_prompt = 0;
   }
   else
      freopen( "con", "r", stdin );

   if ( find_prompt )
      printf( "Enter name to find, Ctrl-Z to end: " );

   symbols = 0;

   while ( gets( s ) )
   {
      ++symbols;
      if ( ( hl = (h_ele *) hash_find( h, s ) ) == NULL )
      {
         ++misses;
         printf( "No %s.\n", s );
      }
      else
         printf( "Name: %s Type: %d.\n", hash_name( h, hl ), hl->name_type );
      if ( find_prompt )
         printf( "Enter name to find, Ctrl-Z to end: " );
   }

   printf( "\nNumber of symbols searched: %d, Number not found: %d\n",
                                        symbols,              misses );

   hash_destroy( h );

} /* main() */

#endif
